
with Text_Io;

with Ada.Strings.Fixed;
use Ada.Strings;

with Ada.Streams.Stream_Io;
use Ada.Streams;

with ada.Task_Identification;
use ada.Task_Identification;
with Gnat.SHA1;
with Gnat.Sockets;
use Gnat;

with Ada.Unchecked_Deallocation;

with Ada.Command_LIne;
use Ada.Command_Line;
with Warborg.Database;
package body Warborg.Uniborg is

   
   procedure Usage is
      
   begin
      Text_Io.Put_Line("usage : " & Command_name);
   end Usage;



   package User_Manager is
      
      
      procedure Initialize(Filename : in String);
      
      function Add_User(Logname : in String;
			Passwd : in Message_Digest) return Boolean;
      

      
      function Check_passwd(Logname : in String;
			    Passwd : in Message_Digest) return Boolean;
      

   end User_Manager;
   
   
   package body User_Manager is
      
      type String_Access is access String;
      Users_Filename : String_Access;
      
      procedure Initialize(Filename : in String) is
      begin
	 Users_Filename := new String ' (Filename);
      end Initialize;
      
      function Add_User(Logname : in String;
			Passwd : in Message_Digest) return Boolean is

	 Users_File : Stream_Io.File_Type;
	 Users_Stream : Stream_Io.Stream_Access;

	 Old, Current : User_Type;

      begin

	 Initialize(Current, Logname);

	 Set_Pass(Current, Passwd);
	 
	 begin
	    Stream_IO.open(users_File, Stream_Io.In_File, Users_Filename.all);
	    
	    users_stream := Stream_IO.Stream(users_File);
	    while not Stream_Io.End_Of_File(Users_File) loop
	       User_Type'read(Users_Stream, old);
	       if Current.Logname = Old.Logname then
		  Stream_IO.Close(Users_File);
		  return False;
	       end if;
	    end loop;
	    Stream_IO.Close(Users_File);
	 exception
	    when Stream_Io.Name_Error =>
	       Stream_IO.create(users_File, Stream_Io.out_File, Users_Filename.all);
	       Stream_IO.Close(Users_File);
	 end;

	 Stream_IO.open(users_File, Stream_Io.append_File, Users_Filename.all);
	 users_stream := Stream_IO.Stream(users_File);
	 User_Type'write(Users_Stream, Current);
	 Stream_IO.Close(Users_File);
	 return True;
      end add_user;


      function Check_passwd(Logname : in String;
			    Passwd : in Message_Digest) return Boolean is
	 Users_File : Stream_Io.File_Type;
	 Users_Stream : Stream_Io.Stream_Access;


	 Old, Current : User_Type;

      begin
	 
	 
	 Initialize(Current, Logname);
	 Set_Pass(Current, Passwd);
	 
	 Stream_IO.open(users_File, Stream_Io.In_File, Users_Filename.all);
	 
	 users_stream := Stream_IO.Stream(users_File);

	 while not Stream_Io.End_Of_File(Users_file) loop
	    User_Type'read(Users_Stream, old);
	    
	    if Current.Logname = Old.Logname then
	       
	       if Check_Pass(Old, passwd) then
		  --if Current.passwd = Old.passwd then

		  Stream_IO.Close(Users_File);

		  return True;

	       else

		  Stream_IO.Close(Users_File);
		  return False;
	       end if;
	    end if;
	 end loop;
	 
	 Stream_IO.Close(Users_File);
	 return False;
      exception
	 when others =>
	    return False;
      end Check_passwd;

      
   end User_Manager;

   
   
   
   Sgbd : Database.Sgbd_Type;
   

   All_Populations : nation_Manager.Vector := Nation_Manager.Empty_Vector;
   
   procedure Initialize_Populations is

      
      use Nation_Manager;
      Population_Filename : constant String := "data/worldcitiespop.txt";
      Error_Count : Integer := 0;
      File : Text_Io.File_Type;
      Population_Ouest : Long_Long_Integer := 0;
      Population_Est : Long_Long_Integer := 0;
      Separator      : constant Character := Character'Val(44);
   begin
      Text_Io.New_Line;
      Text_Io.Put_Line("Initialize users filename");
      Text_Io.Put_Line("Open File " & Population_Filename);
      Text_Io.Open(File, Text_Io.In_File, Population_Filename);
      Text_Io.Put_Line("reading File " & Population_Filename);
      while not Text_Io.End_Of_File(File) loop
	 
	 declare
	    Line   : constant String := Text_Io.Get_Line(File);	   
	    
	    Nation : Nation_Type;
	    
	    Last_Separator : Natural := 0;	    	  
	    
	    Nation_Latitude : Decimal_Latitude_Type := 0.0;
	    Nation_Longitude : Decimal_Longitude_Type := 0.0;
	 begin
	    

	    for I in 1..2 loop	    
	       Last_Separator := Fixed.Index(Line, "" & Separator, Last_Separator+1) + 1;	       
	    end loop;
	    Fixed.Move(Line(Last_Separator..Fixed.Index(Line, "" & Separator, Last_separator) - 1), Nation.Name, Error, left);
	    
	    Last_Separator := 0;
	    
	    for I in 1..5 loop	    
	       Last_Separator := Fixed.Index(Line, "" & Separator, Last_Separator+1)+1;	       	       
	    end loop;
	    
	    
	    
	    Nation_Latitude := Decimal_Latitude_Type'Value(Line(Last_Separator..Fixed.Index(Line, "" & Separator, Last_separator) - 1));	    	    	    	    
	    Last_Separator := Fixed.Index(Line, "" & Separator, Last_Separator+1) + 1;
	    
	    
	    Nation_Longitude := Decimal_Longitude_Type'Value(Line(Last_Separator..Line'last));--Fixed.Index(Line, "" & Separator, Last_separator) - 1)
	    
	    Nation.Gps_Coordonates := To_GPS(Nation_Longitude, Nation_Latitude);
	    --Print(Nation.GPS_Coordonates);
	    
	    
	    
	    
	    Last_Separator := 0;
	    for I in 1..4 loop
	       Last_Separator := Fixed.Index(Line, "" & Separator, Last_Separator+1) + 1;	       	       
	    end loop;
	    
	    
	    
	    begin
	       Nation.Population := Long_Long_Integer'Value(Line(Last_Separator..Fixed.Index(Line, "" & Separator, Last_separator) - 1));	    	    
	       if Nation.Population > 270000 then
		  case Side_Of(Nation.Gps_Coordonates) is
		     when Ouest =>
			Population_Ouest := Population_Ouest + Nation.Population;
			if Population_Ouest < 3_500_000_000 then		       
			   All_Populations := All_Populations & Nation;
			   --Text_Io.Put_Line("Add Ouest ");
			end if;
		     when Est =>		     			
			Population_Est := Population_Est + Nation.Population;
			if Population_Est < 3_500_000_000 then
			   All_Populations := All_Populations & Nation;
			   --Text_Io.Put_Line("Add Est");
			end if;
		  end case;
	       end if;
	    end;
	    
	    
	 exception
	    
	    when Constraint_Error =>
	       Error_Count := Error_Count + 1;
	       
	 end;
	 
      end loop;
      Text_Io.Close(File);
      Text_Io.Put_Line("Error Count = " & Integer'Image(Error_Count));
      Text_Io.Put_Line("Length = " & Nation_Index_Type'Image(Nation_Manager.Last_Index(All_Populations)));

   end Initialize_Populations;

   type Socket_Access is access all Sockets.Socket_Type;


   task type T_Thread_Warborg(Socket : Socket_Access) is
      entry Get_Id(Id : out Task_Id);
   end T_Thread_Warborg;

   type Thread_Warborg_Access is access T_Thread_Warborg;

   type T_Thread_Warborg_Task is
      record
         Id : Task_Id := Null_Task_id;
         thread_Warborg : thread_Warborg_access;
      end record;
   type thread_Warborg_Task_Access is access T_Thread_Warborg_Task;



   procedure Unchecked_Deallocation_thread_Warborg is
      new Ada.Unchecked_Deallocation(T_Thread_Warborg,thread_Warborg_Access);
   procedure Unchecked_Deallocation_thread_Warborg_task is
      new Ada.Unchecked_Deallocation(T_Thread_Warborg_Task,thread_Warborg_Task_Access);
   
   
   
   
   
   use Nation_Manager;      
   
   
   task body T_Thread_Warborg is
      
      Channel  : constant Sockets.Stream_Access := Sockets.stream(socket.all);
      Logname : Logname_Type := (others => Character'Val(32));
      Passwd  : Sha1.Message_Digest;
      
      Command : Command_Type;
      
      Wopr : wopr_Type;
      
   begin
	    
      accept get_id(id : out task_id) do
         id := current_task;
      end get_id;
      loop	 
	 --Text_Io.Put("attend commande");
	 Command := Command_Type'Input(Channel);
	 case Command is
	    when Login =>
	       logname := Logname_Type'Input(Channel);
	       Passwd  := Sha1.Message_Digest'Input(Channel);
	       if User_Manager.Add_User(Logname, Passwd) or User_Manager.Check_Passwd(Logname, Passwd) then
		  --Text_Io.Put_line("envoie login");
		  Boolean'Output(Channel, True);
		  exit;
	       else
		  Boolean'Output(Channel, False);
		  exit;
	       end if;
	    when others =>
	       exit;
	 end case;
      end loop;

      
      loop
	 
	 begin
	    Text_Io.Put_Line("Wait command for " & logname);
	    Command := Command_Type'Input(Channel);
	    
	    case Command is
	       when Wopr_Create =>
		  declare
		     
		     Success : Boolean;
		  begin		     		     		     		     
		     Wopr := Wopr_Type'Input(Channel);
		     Wopr.Destroyed_Missiles := 0;
		     Wopr.Destroyed_Wopr := 0;
		     Wopr.Earned_Missiles := 0;
		     Wopr.Charge := Initial_Charge;
		     Wopr.Populations := All_Populations;
		     --Text_Io.Put_Line("initial length :" & Nation_Index_type'Image(Nation_Manager.Last_Index(Wopr.Populations)));
		     --Text_Io.Put_Line("Adding New wopr in database");
		     Sgbd.Add_Wopr(Logname, Wopr, success);
		     Boolean'Output(Channel, Success);
		     if Success then
			Text_Io.Put_Line("New wopr");			
		     else
			Text_Io.Put_Line("Bad comportment.");
		     end if;
		  end;
	       when Wopr_get =>
		  declare
		     
		     Success : Boolean := False;
		  begin
		     Text_Io.Put_Line("Getting wopr in database");
		     Wopr := Wopr_Type'Input(Channel);
		     Sgbd.Get_Wopr(Logname, Wopr, Success);
		     if Success then
			Text_Io.Put_Line("Getting wopr success");
			Command_Type'output(Channel, Wopr_get);
			
			--Text_Io.Put_Line("after get in database  length :" & natural'Image(Natural (Nation_Manager.Last_Index(Wopr.Populations))));
			Wopr_Type'Output(Channel, Wopr);
			Text_Io.Put_Line("Wopr Sended");

		     else
			Text_Io.Put_Line("Getting wopr failled");
			Command_Type'output(Channel, logout);
		     end if;		     		     		     
		     
		  exception
		     when others =>
			Command_Type'output(Channel, logout);
			Text_Io.Put_Line("                                 Exception in uniborg : Wopr_Get faillure.");
		  end;
	       when Missile_Create =>
		  declare
		     Missile : Missile_Type;
		     Success : Boolean := False;
		  begin

		     Missile := missile_Type'Input(Channel);

		     Sgbd.Add_Missile(Logname, Missile, success);

		     Boolean'Output(Channel, success);
		  end;
	       when Launch_All =>
		  declare
		     
		     Success : Boolean := False;
		  begin
		     -- Je bricole une soluce pour que ça marche, mais c'est pas normal.
		     Boolean'Output(Channel, true);		     		     
		     Text_Io.Put_Line("Uniborg : launch Logname");
		     Sgbd.Launch_all(Logname);
		     Text_Io.Put_Line("Uniborg : launch Logname OK");
		     
		  end;
	       when Enemy_Get =>
		  declare
		     Enemy : Weapon_Manager.Vector := Weapon_Manager.Empty_Vector;		
		  begin
		     Sgbd.Get_Enemy(Logname, enemy);
		     if not Weapon_Manager.Is_Empty(Enemy) then
			Boolean'Output(Channel, True);
			
			Weapon_Manager.Vector'Output(Channel, Enemy);
		     else
			Boolean'Output(Channel, False);
		     end if;
		  end;
	       when others =>
		  null;
	    end case;
	    
	 end;
	 
      end loop;
      
   end T_Thread_Warborg;         
   
   
   
   
   
   
   
   
   
   
   task body Warborg_Server is
      
      Max : constant Positive := 250;
      
      Liste : array(1..max) of Socket_access;
      Liste_Thread_task : array(1..max) of Thread_Warborg_Task_Access;
      Bliste : array(1..max) of Boolean;
      Reads,Write : Sockets.Socket_Set_Type;

      Selector : Sockets.Selector_Type;
      Srv_Adr,Adr : Sockets.Sock_Addr_Type;
      Srv_Socket : Sockets.Socket_Type;
      Status : Sockets.Selector_Status;

      Ptr : Integer := 0;

      function Chercher return Natural is
      begin
         for I in 1..max loop
            if Liste_Thread_Task(I) = null then
               return I;            
            elsif Is_Terminated(Liste_Thread_Task(I).Id) then
               Unchecked_Deallocation_Thread_Warborg(Liste_Thread_Task(I).Thread_Warborg);
               Unchecked_Deallocation_Thread_Warborg_task(Liste_Thread_Task(I));
               Sockets.Close_Socket(Liste(I).all);
               return I;
            end if;
         end loop;
         return 0;
      end Chercher;

      thread_Suivant : Natural := 0;



   begin
      
      if Argument_Count /=0 then
	 Usage;
	 raise Program_Error;
      end if;
      
      User_Manager.Initialize("users.bin");
      
      
      Initialize_Populations;

      
      
      

      for I in 1..max loop
         liste(I) := new Sockets.Socket_Type;
      end loop;


      Bliste := (others => False);

      Srv_Adr.Addr := Sockets.Any_Inet_Addr;
      Srv_Adr.Port := 5872;
      Sockets.Create_Socket(Srv_Socket);

      Sockets.Bind_Socket(Srv_Socket,Srv_Adr);
      Sockets.Listen_Socket(Srv_Socket);
      Sockets.Create_Selector(Selector);
      Sockets.Set(Write,Srv_Socket);

      
      
      loop
	 begin
	    Sockets.Empty(Reads);
	    
	    Sockets.Set(Reads, Srv_Socket);
	    
	    for I in 1..max loop
	       if Bliste(I) then		  
		  Sockets.Set(Write,Liste(I).all);
	       end if;
	    end loop;
	    
	    Sockets.Check_Selector(Selector, Reads, Write, Status);

	    case Status is
	       when Sockets.Completed =>

		  if Sockets.Is_Set(Reads, Srv_Socket) or Sockets.Is_Set(Write, Srv_Socket)  then
		     Thread_Suivant := Chercher;

		     if Thread_Suivant /= 0 then
			Ptr := Thread_Suivant;

			Sockets.Accept_Socket(Srv_Socket,Liste(Ptr).all,Adr);
			Bliste(Ptr) := True;
			
		     end if;
		  else
		     
		     for I in 1..max loop


			if bliste(I) then

			   if Sockets.Is_Set(Reads,Liste(I).all) or Sockets.Is_Set(Write,Liste(I).all) then
			      Liste_Thread_Task(I) := new T_Thread_Warborg_Task;
			      
			      Liste_Thread_Task(I).Thread_Warborg := new T_Thread_Warborg(Liste(i));

			      Liste_Thread_Task(i).Thread_Warborg.Get_Id(Liste_Thread_Task(i).Id);
			      

			      Bliste(i) := false;

			   end if;
			end if;
		     end loop;
		  end if;
		  delay 1.0;
	       when Sockets.Expired =>
		  Text_Io.Put_Line("expired");
	       when Sockets.Aborted =>
		  Text_Io.Put_Line("aborted");
	    end case;
	    
	 exception
	    when others =>
	       Text_Io.Put_Line("Error Warborg server...");
	       Sockets.Create_Selector(Selector);
	       Sockets.Set(Write,Srv_Socket);
	 end;
      end loop;
      
   end Warborg_server;


   
   

end Warborg.Uniborg;
